from glob import iglob
from importlib import import_module
from os import path as pth
from typing import Tuple

from .types import CmdMap, CommandModule, MetaMap, MetaModule, NestedCmdMap


def _get_command_module(name: str) -> CommandModule | None:
    cmd_mod = import_module(name).__dict__
    run_func = cmd_mod.get('run')
    if run_func:
        return CommandModule(run=run_func)
    return None


def _get_meta_module(name: str) -> MetaModule:
    meta_mod = import_module(name).__dict__
    return MetaModule(
        meta=meta_mod['meta'],
        add_arguments=meta_mod.get('add_arguments'),
    )


def get_command_modules(root: str, commands_module: str) -> CmdMap:
    """Create a dictionary of a file name to its module.

    These modules should define a `run` function.

    Args:
        root: Absolute path of the directory containing the __main__.py module
            that activates the cli.
        commands_module: The full module resolution. For instance
            `m.cli.commands`.

    Returns:
        A dictionary of command modules.
    """
    dir_name = '/'.join(commands_module.split('.')[1:])
    mod_names = list(iglob(f'{root}/{dir_name}/*.py'))
    mod = {}
    for mod_name in mod_names:
        name = pth.split(mod_name)[1][:-3]
        cmd_mod = _get_command_module(f'{commands_module}.{name}')
        if cmd_mod:
            mod[name] = cmd_mod
    return mod


def get_cli_command_modules(file_path: str) -> Tuple[NestedCmdMap, MetaMap]:
    """Return a dictionary containing the commands and subcommands for the cli.

    `file_path` is expected to be the absolute path to the __main__.py file.
    Another restriction is that the __main__.py file must have the
    `cli.commands` module as its sibling.

    Args:
        file_path: Absolute path to the `__main__.py` file.

    Returns:
        A tuple with a map containing all the modules that have the commands
        for the cli and a map with the meta data related to each command.
    """
    root = pth.split(pth.abspath(file_path))[0]
    main_mod = pth.split(root)[1]
    cli_root = f'{main_mod}.cli'
    root_cmd = get_command_modules(root, f'{cli_root}.commands')
    mod: NestedCmdMap = {}
    meta: MetaMap = {}
    for key, cmd_mod in root_cmd.items():
        mod[key] = cmd_mod
    meta['_root'] = _get_meta_module(f'{cli_root}.commands')
    subcommands = list(iglob(f'{root}/cli/commands/*'))
    for cmd_name in subcommands:
        if cmd_name.endswith('.py') or cmd_name.endswith('__'):
            continue
        name = pth.split(cmd_name)[1]
        mod[name] = get_command_modules(root, f'{cli_root}.commands.{name}')
        meta[name] = _get_meta_module(f'{cli_root}.commands.{name}')
    return mod, meta



import argparse
import sys
from typing import Any, Callable, cast

from m.core import Issue, OneOf
from m.core.fp import is_bad
from m.log import Logger

from ..core.io import env
from .engine.misc import params_count
from .engine.sys import get_cli_command_modules
from .engine.types import CmdMap, CommandModule, MetaMap, NestedCmdMap
from .handlers import create_issue_handler, create_json_handler
from .validators import validate_non_empty_str

logger = Logger('m.cli')
cli_global_option = Callable[[argparse.ArgumentParser], argparse.Action]


default_result_handler = create_json_handler(pretty=False)
default_issue_handler = create_issue_handler(use_warning=False)


def _main_parser(
    mod: NestedCmdMap,
    meta: MetaMap,
    add_args: Callable[[argparse.ArgumentParser], None] | None = None,
) -> argparse.Namespace:
    main_meta = meta['_root'].meta
    raw = argparse.RawTextHelpFormatter
    # NOTE: In the future we will need to extend from this class to be able to
    #   override the error method to be able to print CI environment messages.
    argp = argparse.ArgumentParser(
        formatter_class=raw,
        description=main_meta['description'] if main_meta else '...',
    )
    if add_args:
        add_args(argp)
    subp = argp.add_subparsers(
        title='commands',
        dest='command_name',
        required=True,
        help='additional help',
        metavar='<command>',
    )
    names = sorted(mod.keys())
    for name in names:
        if isinstance(mod[name], dict):
            meta_mod = meta[name]
            meta_dict = meta_mod.meta
            parser = subp.add_parser(
                name,
                help=meta_dict['help'],
                formatter_class=raw,
                description=meta_dict['description'],
            )
            if meta_mod.add_arguments:
                meta_mod.add_arguments(parser)
            subsubp = parser.add_subparsers(
                title='commands',
                dest='subcommand_name',
                required=True,
                help='additional help',
                metavar='<command>',
            )
            sub_mod = cast(CmdMap, mod[name])
            for subname in sorted(sub_mod.keys()):
                run_func = sub_mod[subname].run
                if params_count(run_func) == 2:
                    run_func(None, subsubp)
        else:
            mod_inst = cast(CommandModule, mod[name])
            run_func = mod_inst.run
            if params_count(run_func) == 2:
                run_func(None, subp)
    return argp.parse_args()


def run_cli(
    file_path: str,
    add_args: Callable[[argparse.ArgumentParser], None] | None = None,
) -> None:
    """Run the cli application.

    usage::

        def add_args(argp):
            argp.add_argument(...)
        def main():
            run_cli(__file__, add_args)

    We only need `add_args` if we need to gain access to the
    `argparse.ArgumentParser` instance.

    Args:
        file_path: The full name of the file: __file__.
        add_args: Optional callback to gain access to the ArgumentParser.
    """
    mod, meta = get_cli_command_modules(file_path)
    arg = _main_parser(mod, meta, add_args)

    run_func = None

    if hasattr(arg, 'subcommand_name'):
        sub_mod = cast(CmdMap, mod[arg.command_name])
        run_func = sub_mod[arg.subcommand_name].run
    else:
        run_func = cast(CommandModule, mod[arg.command_name]).run

    len_run_args = params_count(run_func)
    run_args = [arg, None][:len_run_args]
    exit_code = 0
    try:
        # mypy is having a hard time figuring out the type of run_args
        exit_code = run_func(*run_args)  # type:ignore[arg-type]
    except Exception as ex:
        default_issue_handler(
            Issue('unknown cli run function exception', cause=ex),
        )
        exit_code = 5
    sys.exit(exit_code)


def run_main(
    callback: Callable[[], OneOf[Issue, Any]],
    result_handler: Callable[[Any], None] = default_result_handler,
    issue_handler: Callable[[Issue], None] = default_issue_handler,
) -> int:
    """Run the callback and print the returned value.

    To change how the result or an issue should be display provide the optional
    arguments `handle_result` and `handle_issue`. For instance, to display the
    raw value provide the `print` function.

    Args:
        callback: A function that returns a `OneOf`.
        result_handler: A function that takes in the Good result.
        issue_handler: A function that takes in the Issue.

    Returns:
        0 if the callback is a `Good` result otherwise return 1.
    """
    res = None
    try:
        res = callback()
    except Exception as ex:
        issue_handler(Issue('unknown caught exception', cause=ex))
        return 2
    if is_bad(res):
        problem = res.value
        if isinstance(problem, Issue):
            issue_handler(problem)
        else:
            issue_handler(Issue('non-issue exception', cause=problem))
        return 1
    result_handler(res.value)
    return 0


def cli_integration_token(
    integration: str,
    env_var: str,
) -> cli_global_option:
    """Return a function that takes in a parser.

    This generated function registers a token argument in the parser
    which looks for its value in the environment variables.

    Args:
        integration: The name of the integration.
        env_var: The environment variable name.

    Returns:
        A function to add arguments to an argparse parser.
    """
    return lambda parser: parser.add_argument(
        '-t',
        '--token',
        type=validate_non_empty_str,
        default=env(env_var),
        help=f'{integration} access token (default: env.{env_var})',
    )



from .cli import cli_integration_token, run_cli, run_main
from .engine.argparse import command
from .engine.types import FuncArgs, add_arg
from .handlers import (
    create_issue_handler,
    create_json_handler,
    create_yaml_handler,
)
from .validators import validate_json_payload, validate_payload

# using as barrel file to export convenience functions
__all__ = (  # noqa: WPS410
    'cli_integration_token',
    'create_issue_handler',
    'create_json_handler',
    'create_yaml_handler',
    'command',
    'add_arg',
    'validate_json_payload',
    'validate_payload',
    'run_main',
    'run_cli',
    'FuncArgs',
)



import argparse as ap
from dataclasses import dataclass
from typing import Any, Callable, TypeVar

from pydantic import BaseModel


@dataclass
class CommandInputs:
    """Inputs to command decorator."""

    name: str
    help: str
    model: type[BaseModel]


@dataclass
class FuncArgs:
    """Stores function arguments."""

    args: list[Any]
    kwargs: dict[str, Any]


def add_arg(*args, **kwargs) -> FuncArgs:
    """Wrap FuncArgs arguments in a function.

    Args:
        args: The arguments to argparse add_arguments.
        kwargs: The keyword arguments to argparse add arguments.

    Returns:
        A FuncArgs instance.
    """
    return FuncArgs(args=list(args), kwargs=kwargs)


@dataclass
class CommandModule:
    """Container to store the run function from a "command" module."""

    run: Callable[
        [ap.Namespace | None, ap._SubParsersAction | None],  # noqa: WPS437
        int,
    ]


@dataclass
class MetaModule:
    """Container to store a metadata dictionary from a "meta" module."""

    meta: dict[str, str]
    add_arguments: Callable[[ap.ArgumentParser], None] | None


MetaMap = dict[str, MetaModule]
CmdMap = dict[str, CommandModule]
NestedCmdMap = dict[str, CommandModule | CmdMap]
MISSING = TypeVar('MISSING')
AnyMap = dict[str, Any]
